package holeg.ui.model;

import java.util.ArrayList;
import java.util.HashMap;

import holeg.model.AbstractCanvasObject;
import holeg.model.Edge;
import holeg.model.GroupNode;
import holeg.model.HolonObject;
import holeg.model.HolonSwitch;
import holeg.model.Node;

/**
 * For DecoratedState And VisualRepresentationalState
 * @author Tom
 *
 */
public class MinimumModel {
	
	
	
	
	private ArrayList<HolonObject> holonObjectList = new ArrayList<>();
	private ArrayList<Edge> edgeList = new ArrayList<>();
	private ArrayList<Node> nodeList = new ArrayList<>();
	private ArrayList<HolonSwitch> switchList = new ArrayList<>();
	//-->Only for Visual:
	private ArrayList<GroupNode> uppderNodeList = new ArrayList<>();
	private HashMap<AbstractCanvasObject, GroupNode> inGroupObjects = new HashMap<>();
	HashMap<Edge, ArrayList<GroupNode>> inGroupEdges = new HashMap<>();
	//<--
	
	
	
	
	
	public HashMap<Edge, ArrayList<GroupNode>> getInGroupEdges() {
		return inGroupEdges;
	}

	public MinimumModel(ArrayList<AbstractCanvasObject> abstractObjectList, ArrayList<Edge> edgeList, int timestep) {// Constructor because of old Model
		for (AbstractCanvasObject aCps : abstractObjectList) {
			if (aCps instanceof HolonObject hObject) {
				holonObjectList.add(hObject);
			}
			else if (aCps instanceof Node node) nodeList.add(node);
			else if (aCps instanceof HolonSwitch sw) switchList.add(sw);
			else if(aCps instanceof GroupNode groupnode) {
				addGroupNodeObjects(groupnode, timestep);
				uppderNodeList.add(groupnode);
			}
		}
		for (Edge edge : edgeList) {
			this.edgeList.add(edge);
			AbstractCanvasObject objectA = edge.getA();
			AbstractCanvasObject objectB = edge.getB();
			if(inGroupObjects.containsKey(objectA) || inGroupObjects.containsKey(objectB)){
				ArrayList<GroupNode> list = new ArrayList<GroupNode>(); 
				if(inGroupObjects.containsKey(objectA)) {
					list.add(inGroupObjects.get(objectA));
				}
				if(inGroupObjects.containsKey(objectB)) {
					list.add(inGroupObjects.get(objectB));					
				}
				inGroupEdges.put(edge, list);
			}
		}
		//Calculate Energy of HolonObjects
		holonObjectList.forEach(obj -> obj.calculateEnergy(timestep));
	}

	private void addGroupNodeObjects(GroupNode aUpperNode, int timestep) {
		for(AbstractCanvasObject aCps : aUpperNode.getNodes()) {
			if (aCps instanceof HolonObject hObject) {
				holonObjectList.add(hObject);
			}
			else if (aCps instanceof Node node) nodeList.add(node);
			else if (aCps instanceof HolonSwitch sw) switchList.add(sw);
			else if(aCps instanceof GroupNode groupnode) {
				addGroupNodeObjects(groupnode, timestep);
				uppderNodeList.add(groupnode);
			}
			inGroupObjects.put(aCps, aUpperNode);
		}
	}
	
	public ArrayList<HolonObject> getHolonObjectList() {
		return holonObjectList;
	}
	public void setHolonObjectList(ArrayList<HolonObject> holonObjectList) {
		this.holonObjectList = holonObjectList;
	}
	public ArrayList<Edge> getEdgeList() {
		return edgeList;
	}
	public void setEdgeList(ArrayList<Edge> cableList) {
		this.edgeList = cableList;
	}
	public ArrayList<Node> getNodeList() {
		return nodeList;
	}
	public void setNodeList(ArrayList<Node> nodeList) {
		this.nodeList = nodeList;
	}
	public ArrayList<HolonSwitch> getSwitchList() {
		return switchList;
	}
	public void setSwitchList(ArrayList<HolonSwitch> switchList) {
		this.switchList = switchList;
	}

	public ArrayList<GroupNode> getUppderNodeList() {
		return uppderNodeList;
	}

	public HashMap<AbstractCanvasObject, GroupNode> getInGroupObjects() {
		return inGroupObjects;
	}

	
}
